# SPDX-License-Identifier: Apache-2.0
# The OpenSearch Contributors require contributions made to
# this file be licensed under the Apache-2.0 license or a
# compatible open source license.
# Any modifications Copyright OpenSearch Contributors. See
# GitHub history for details.

BANDWIDTH = 16.0
NMF_MAX_ITER = 300
NMF_TOL = 1e-4
DT_PVAL = 0.5
OMP_TOL = 0.8
