# SPDX-License-Identifier: Apache-2.0
# The OpenSearch Contributors require contributions made to
# this file be licensed under the Apache-2.0 license or a
# compatible open source license.
# Any modifications Copyright OpenSearch Contributors. See
# GitHub history for details.


#  Licensed to Elasticsearch B.V. under one or more contributor
#  license agreements. See the NOTICE file distributed with
#  this work for additional information regarding copyright
#  ownership. Elasticsearch B.V. licenses this file to you under
#  the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.

# TODO fill out this information
__title__ = "opensearch_py_ml"
__description__ = "Python Client and Toolkit for DataFrames, Big Data, Machine Learning and ETL in OpenSearch"
__url__ = "https://github.com/opensearch-project/opensearch-py-ml"
__version__ = "1.3.0"
__author__ = "Dhrubo Saha"
__author_email__ = "dhrubo@amazon.com, "
__maintainer__ = "Dhrubo Saha, Yaliang Wu, Alexander Greaves-Tunnell, Sicheng Song, Jing Zhang, Bhavana Goud Ramaram, Yerzhaisang Taskali, Nathalie Jonathan"
__maintainer_email__ = "dhrubo@amazon.com, ylwu@amazon.com, greaa@amazon.com, seasonsg@amazon.com, jngz@amazon.com, rbhavna@amazon.com, tasqali1697@gmail.com, nathhjo@amazon.com"
